/*=========================================================================

  Program:   Insight Segmentation & Registration Toolkit
  Module:    $RCSfile: itkNaryFunctorImageFilter.txx,v $
  Language:  C++
  Date:      $Date: 2004/05/03 17:00:38 $
  Version:   $Revision: 1.17 $

  Copyright (c) Insight Software Consortium. All rights reserved.
  See ITKCopyright.txt or http://www.itk.org/HTML/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even 
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef _itkNaryFunctorImageFilter_txx
#define _itkNaryFunctorImageFilter_txx

#include "itkNaryFunctorImageFilter.h"
#include "itkImageRegionIterator.h"
#include "itkProgressReporter.h"

namespace itk
{

/**
 * Constructor
 */
template <class TInputImage, class TOutputImage, class TFunction >
NaryFunctorImageFilter<TInputImage,TOutputImage,TFunction>
::NaryFunctorImageFilter()
{
  // This number will be incremented each time an image
  // is added over the two minimum required
  this->SetNumberOfRequiredInputs( 2 );
  this->InPlaceOff();
}


/**
 * ThreadedGenerateData Performs the pixel-wise addition
 */
template <class TInputImage, class TOutputImage, class TFunction>
void
NaryFunctorImageFilter<TInputImage, TOutputImage, TFunction>
::ThreadedGenerateData( const OutputImageRegionType &outputRegionForThread,
                        int threadId)
{

  const unsigned int numberOfInputImages = 
    static_cast< unsigned int >( this->GetNumberOfInputs() );
  
  OutputImagePointer outputPtr = this->GetOutput(0);
  ImageRegionIterator<TOutputImage> outputIt(outputPtr, outputRegionForThread);


  // Clear the content of the output
  outputIt.GoToBegin();
  while( !outputIt.IsAtEnd() )
    {
    outputIt.Set( itk::NumericTraits< OutputImagePixelType >::Zero );
    ++outputIt;
    }
   
  // support progress methods/callbacks.
  // count the number of inputs that are non-null
  unsigned int numberOfValidInputImages = 0;
  for (unsigned int i=0; i < numberOfInputImages; ++i)
    {
    InputImagePointer inputPtr =
      dynamic_cast<TInputImage*>( ProcessObject::GetInput( i ) );

    if (inputPtr)
      {
      numberOfValidInputImages++;
      }
    }
  ProgressReporter progress(this, threadId,
                            numberOfValidInputImages
                            *outputRegionForThread.GetNumberOfPixels());
      
  for(unsigned int inputNumber=0;
      inputNumber < numberOfInputImages; inputNumber++ )
    {
    // We use dynamic_cast since inputs are stored as DataObjects.  
    InputImagePointer inputPtr = 
      dynamic_cast<TInputImage*>( ProcessObject::GetInput( inputNumber ) );

    if (inputPtr)
      {
      ImageRegionIterator<TInputImage> inputIt(inputPtr,outputRegionForThread);
      
      inputIt.GoToBegin();
      outputIt.GoToBegin();
      while( !inputIt.IsAtEnd() ) 
        {
        outputIt.Set( m_Functor( outputIt.Get(), inputIt.Get() ) );
        ++inputIt;
        ++outputIt;
        progress.CompletedPixel();
        }
      }
    }
}

} // end namespace itk

#endif


